# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom git mergetool for renku metadata."""

from pathlib import Path

from pydantic import ConfigDict, validate_call

from renku.command.command_builder.command import Command
from renku.domain_model.project_context import project_context


def mergetool_command():
    """Command to move or rename a file, a directory, or a symlink."""
    return Command().command(_mergetool).require_migration().with_database()


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def _mergetool(local: Path, remote: Path, base: Path) -> None:
    """Merge renku metadata files.

    Args:
        local(Path): The file to merge from the local branch.
        remote(Path): The file to merge from the remote branch.
        base(Path): Path to common base of branches to be merged.
    """
    from renku.infrastructure.git_merger import GitMerger

    merger = GitMerger()

    merger.merge(local, remote, base)


def mergetool_install_command():
    """Command to setup renku as a custom git merge tool."""
    return Command().command(setup_mergetool)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def setup_mergetool(with_attributes: bool = True):
    """Setup renku custom mergetool."""
    repository = project_context.repository

    with repository.get_configuration(writable=True) as config_writer:
        config_writer.set_value('merge "renkumerge"', "name", "Renku merge driver")
        config_writer.set_value('merge "renkumerge"', "driver", "renku mergetool merge %O %A %B")
        config_writer.set_value('merge "renkumerge"', "trustExitCode", "true")
        config_writer.set_value('merge "renkumerge"', "recursive", "binary")

    if not with_attributes:
        return

    attributes_path = project_context.path / ".gitattributes"
    pattern_string = ".renku/metadata/**    merge=renkumerge\n"

    if attributes_path.exists():
        with open(attributes_path) as f:
            content = f.readlines()
            if pattern_string in content:
                return

    with open(attributes_path, "a") as f:
        f.write(pattern_string)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template use cases."""
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import click
from pydantic import ConfigDict, validate_call

from renku.command.command_builder.command import inject
from renku.command.view_model.template import TemplateChangeViewModel, TemplateViewModel
from renku.core import errors
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.template.template import (
    FileAction,
    RepositoryTemplates,
    TemplateAction,
    copy_template_to_project,
    fetch_templates_source,
    get_file_actions,
    has_template_checksum,
    read_template_checksum,
    set_template_parameters,
    write_template_checksum,
)
from renku.core.util import communication
from renku.core.util.metadata import is_renku_project, replace_renku_version_in_dockerfile
from renku.core.util.tabulate import tabulate
from renku.domain_model.project import Project
from renku.domain_model.project_context import project_context
from renku.domain_model.template import (
    RenderedTemplate,
    Template,
    TemplateMetadata,
    TemplatesSource,
    calculate_dockerfile_checksum,
)
from renku.infrastructure.repository import DiffLineChangeType, Repository


def update_dockerfile_checksum(new_checksum: str):
    """Update ``Dockerfile`` template checksum if possible."""
    if not project_context.dockerfile_path.exists():
        raise errors.DockerfileUpdateError("Project doesn't have a Dockerfile")
    if is_dockerfile_updated_by_user():
        raise errors.DockerfileUpdateError("Cannot update Dockerfile checksum because it was updated by the user")

    checksums = read_template_checksum()
    checksums["Dockerfile"] = new_checksum
    write_template_checksum(checksums)


def does_dockerfile_contain_only_version_change() -> bool:
    """Return True if Dockerfile only contains Renku version changes."""
    commits = list(project_context.repository.iterate_commits(project_context.dockerfile_path))
    # NOTE: Don't include the first commit that added the Dockerfile
    for commit in commits[:-1]:
        changes = commit.get_changes(project_context.dockerfile_path, patch=True)
        if not changes:
            continue
        diff = changes[0].diff
        # NOTE: Check the Dockerfile change only includes adding and removing a Renku version line
        if (
            len(diff) != 2
            or {c.change_type for c in diff} != {DiffLineChangeType.ADDED, DiffLineChangeType.DELETED}
            or any("ARG RENKU_VERSION=" not in c.text for c in diff)
        ):
            return False

    return True


def is_dockerfile_updated_by_user() -> bool:
    """Return if user modified the ``Dockerfile``."""
    dockerfile = project_context.dockerfile_path

    if not has_template_checksum() or not dockerfile.exists():
        return False

    original_checksum = read_template_checksum().get("Dockerfile")
    current_checksum = calculate_dockerfile_checksum(dockerfile=dockerfile)

    if original_checksum == current_checksum:  # Dockerfile was never updated
        return False

    # NOTE: Check if original Dockerfile has the same checksum as the time when the template was set/updated
    metadata = json.loads(project_context.project.template_metadata.metadata)
    original_renku_version = metadata.get("__renku_version__")

    original_dockerfile_content = replace_renku_version_in_dockerfile(dockerfile.read_text(), original_renku_version)
    original_calculated_checksum = calculate_dockerfile_checksum(dockerfile_content=original_dockerfile_content)

    if original_checksum == original_calculated_checksum:
        return False

    return False if does_dockerfile_contain_only_version_change() else True


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def list_templates(source: Optional[str], reference: Optional[str]) -> List[TemplateViewModel]:
    """Return available templates from a source."""
    templates_source = fetch_templates_source(source=source, reference=reference)

    return [TemplateViewModel.from_template(t) for t in templates_source.templates]


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def show_template(source: Optional[str], reference: Optional[str], id: Optional[str]) -> TemplateViewModel:
    """Show template details."""
    if source or id:
        templates_source = fetch_templates_source(source=source, reference=reference)
        template = templates_source.get_template(id=id, reference=None)
    elif is_renku_project():
        metadata = TemplateMetadata.from_project(project=project_context.project)

        templates_source = fetch_templates_source(source=metadata.source, reference=metadata.reference)
        id = metadata.id
        template = templates_source.get_template(id=id, reference=None)
    else:
        raise errors.ParameterError("No Renku project found")

    if template is None:
        raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available")

    return TemplateViewModel.from_template(template)


def check_for_template_update(project: Optional[Project]) -> Tuple[bool, bool, Optional[str], Optional[str]]:
    """Check if the project can be updated to a newer version of the project template."""
    metadata = TemplateMetadata.from_project(project=project)

    templates_source = fetch_templates_source(source=metadata.source, reference=metadata.reference)
    update_available, latest_reference = templates_source.is_update_available(
        id=metadata.id, reference=metadata.reference, version=metadata.version
    )

    return update_available, metadata.allow_update, metadata.reference, latest_reference


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def set_template(
    source: Optional[str],
    reference: Optional[str],
    id: Optional[str],
    force: bool,
    interactive: bool,
    input_parameters: Optional[Dict[str, str]],
    dry_run: bool,
) -> TemplateChangeViewModel:
    """Set template for a project."""
    project = project_context.project

    if project.template_metadata.template_source and not force:
        raise errors.TemplateUpdateError("Project already has a template: To set a template use '-f/--force' flag")

    templates_source = fetch_templates_source(source=source, reference=reference)

    template = select_template(templates_source, id=id)

    if template is None:
        raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available")

    rendered_template, actions = _set_or_update_project_from_template(
        templates_source=templates_source,
        reference=template.reference,
        id=template.id,
        interactive=interactive,
        dry_run=dry_run,
        template_action=TemplateAction.SET,
        input_parameters=input_parameters,
    )

    return TemplateChangeViewModel.from_template(template=rendered_template, actions=actions)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def update_template(force: bool, interactive: bool, dry_run: bool) -> Optional[TemplateChangeViewModel]:
    """Update project's template if possible. Return corresponding viewmodel if updated."""
    template_metadata = TemplateMetadata.from_project(project=project_context.project)

    if not template_metadata.source:
        raise errors.TemplateUpdateError("Project doesn't have a template: Use 'renku template set'")
    if not has_template_checksum() and not interactive:
        raise errors.TemplateUpdateError("Required template metadata doesn't exist: Use '-i/--interactive' flag")

    if not template_metadata.allow_update and not force:
        raise errors.TemplateUpdateError(
            "Update is not allowed for this template. You can still update it using '-f/--force' flag but it might "
            "break your project"
        )

    try:
        templates_source = fetch_templates_source(
            source=template_metadata.source, reference=template_metadata.reference
        )
    except errors.TemplateMissingReferenceError as e:
        message = f"{str(e)}; You can still manually update the template and set a difference reference."
        raise errors.TemplateUpdateError(message)
    except errors.InvalidTemplateError:
        raise errors.TemplateUpdateError("Template cannot be fetched.")

    update_available, latest_reference = templates_source.is_update_available(
        id=template_metadata.id, reference=template_metadata.reference, version=template_metadata.version
    )

    if not update_available:
        return None

    rendered_template, actions = _set_or_update_project_from_template(
        templates_source=templates_source,
        reference=latest_reference,
        id=template_metadata.id,
        interactive=interactive,
        dry_run=dry_run,
        template_action=TemplateAction.UPDATE,
        input_parameters=None,
    )

    return TemplateChangeViewModel.from_template(
        template=rendered_template, actions=actions, old_id=template_metadata.id
    )


@inject.autoparams("project_gateway")
def _set_or_update_project_from_template(
    templates_source: TemplatesSource,
    reference: str,
    id: str,
    interactive,
    dry_run: bool,
    template_action: TemplateAction,
    input_parameters,
    project_gateway: IProjectGateway,
) -> Tuple[RenderedTemplate, Dict[str, FileAction]]:
    """Update project files and metadata from a template."""
    if interactive and not communication.has_prompt():
        raise errors.ParameterError("Cannot use interactive mode with no prompt")

    input_parameters = input_parameters or {}

    project = project_gateway.get_project()

    template = templates_source.get_template(id=id, reference=reference)

    if template is None:
        raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available")

    template_metadata = TemplateMetadata.from_project(project=project_context.project)
    template_metadata.update(template=template)

    if not dry_run:
        set_template_parameters(
            template=template,
            template_metadata=template_metadata,
            input_parameters=input_parameters,
            interactive=interactive,
        )

    rendered_template = template.render(metadata=template_metadata)
    actions = get_file_actions(
        rendered_template=rendered_template, template_action=template_action, interactive=interactive and not dry_run
    )

    if not dry_run:
        copy_template_to_project(rendered_template=rendered_template, project=project, actions=actions)
        project_gateway.update_project(project)

    return rendered_template, actions


def select_template(templates_source: TemplatesSource, id: Optional[str] = None) -> Optional[Template]:
    """Select a template from a template source."""

    def prompt_to_select_template():
        if not communication.has_prompt():
            raise errors.InvalidTemplateError("Cannot select a template")

        class Selection(NamedTuple):
            number: int
            id: str

        templates = [Selection(number=i, id=t.id) for i, t in enumerate(templates_source.templates, start=1)]
        tables = tabulate(templates, headers=["number", "id"])

        message = f"{tables}\nPlease choose a template by typing its number"

        template_index = communication.prompt(
            msg=message, type=click.IntRange(1, len(templates_source.templates)), show_default=False, show_choices=False
        )
        return templates_source.templates[template_index - 1]

    if id:
        try:
            return templates_source.get_template(id=id, reference=None)
        except errors.TemplateNotFoundError:
            raise errors.TemplateNotFoundError(f"The template with id '{id}' is not available")
    elif len(templates_source.templates) == 1:
        return templates_source.templates[0]

    return prompt_to_select_template()


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def validate_templates(
    source: Optional[str] = None, reference: Optional[str] = None
) -> Dict[str, Union[str, Dict[str, List[str]]]]:
    """Validate a template repository.

    Args:
        source(str, optional): Remote repository URL to clone and check (Default value = None).
        reference(str, optional): Git commit/branch/tag to check (Default value = None).
    Returns:
        Dict[str, Union[str, Dict[str, List[str]]]]: Dictionary containing errors and warnings for manifest and
            templates, along with a ``valid`` field telling if all checks passed.
    """

    if source is not None:
        path = Path(tempfile.mkdtemp())
        repo = Repository.clone_from(path=path, url=source)
        repo.checkout(reference=reference)
    else:
        path = Path(os.getcwd())
        repo = Repository(path=path)

        if reference is not None:
            path = Path(tempfile.mkdtemp())
            repo.create_worktree(path, reference=reference)
            repo = Repository(path=path)

    version = repo.head.commit.hexsha

    result: Dict[str, Any] = {"manifest": None, "templates": {}, "warnings": [], "valid": True}

    try:
        template_source = RepositoryTemplates(
            path=path, source=path, reference="", version=version, repository=repo, skip_validation=True
        )
        result["warnings"] = template_source.manifest.validate(manifest_only=True)
    except errors.InvalidTemplateError as e:
        result["manifest"] = e.args[0] if e.args else str(e)
        result["valid"] = False
        return result

    for template in template_source.manifest.templates:
        template.templates_source = template_source
        issues = template.validate(skip_files=False, raise_errors=False)
        if issues:
            result["templates"][template.id] = issues
            result["valid"] = False

    return result



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Move or rename a file, a directory, or a symlink.

Description
~~~~~~~~~~~

Moving a file that belongs to a dataset will update its metadata to include its
new path and commit. Moreover, tracking information in an external storage
(e.g. Git LFS) will be updated.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.move:move
   :prog: renku mv
   :nested: full

Examples
~~~~~~~~

Move operation fails if a destination already exists in the repo;
use ``--force`` flag to overwrite them.

If you want to move files to another dataset use ``--to-dataset`` along with
destination's dataset name. This removes source paths from all datasets'
metadata that include them (if any) and adds them to the destination's dataset
metadata.

The following command moves ``data/src`` and ``README`` to ``data/dst``
directory and adds them to ``target-dataset``'s metadata. If the source files
belong to one or more datasets then they will be removed from their metadata.

.. code-block:: console

    $ renku mv data/src README data/dst --to-dataset target-dataset

.. cheatsheet::
   :group: Misc
   :command: $ renku mv <path>... <destination>
   :description: Safely move files within a project.
   :target: rp

"""

import click

from renku.ui.cli.utils.callback import ClickCallback


@click.command(name="mv")
@click.argument("sources", type=click.Path(exists=True), nargs=-1)
@click.argument("destination", type=click.Path(), nargs=1)
@click.option("-f", "--force", is_flag=True, help="Override existing files.")
@click.option("-v", "--verbose", is_flag=True, help="Show move sources and destinations.")
@click.option("--to-dataset", type=str, default=None, nargs=1, help="A target dataset to move files to.")
def move(sources, destination, force, verbose, to_dataset):
    """Move files and check repository for potential problems."""
    from renku.command.move import move_command

    communicator = ClickCallback()
    move_command().with_communicator(communicator).build().execute(
        sources=sources, destination=destination, force=force, verbose=verbose, to_dataset=to_dataset
    )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent an annotation for a workflow."""

import copy
from uuid import uuid4


class Annotation:
    """Represents a custom annotation for a research object."""

    def __init__(self, *, id: str, body=None, source=None):
        self.id = id
        self.body = body
        self.source = source

    def copy(self):
        """Return a copy of this annotation."""
        return copy.deepcopy(self)

    @staticmethod
    def generate_id():
        """Generate an id for an annotation."""
        return f"/annotations/{uuid4().hex}"



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repository cache interface."""

from abc import ABC
from typing import Optional

from renku.ui.service.cache import ServiceCache
from renku.ui.service.cache.models.project import Project
from renku.ui.service.cache.models.user import User


class IRepositoryCache(ABC):
    """Interface for repository cache manager."""

    def get(
        self, cache: ServiceCache, git_url: str, branch: Optional[str], user: User, shallow: bool = True
    ) -> Project:
        """Get a project from cache (clone if necessary)."""
        raise NotImplementedError()

    def evict(self, project: Project):
        """Evict a project from cache."""
        raise NotImplementedError()

    def evict_expired(self):
        """Evict expired projects from cache."""
        raise NotImplementedError()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku plugin implementations."""

from typing import TYPE_CHECKING, List, Type

from renku.core.dataset.providers.azure import AzureProvider
from renku.core.dataset.providers.dataverse import DataverseProvider
from renku.core.dataset.providers.external import ExternalProvider
from renku.core.dataset.providers.git import GitProvider
from renku.core.dataset.providers.local import LocalProvider
from renku.core.dataset.providers.olos import OLOSProvider
from renku.core.dataset.providers.renku import RenkuProvider
from renku.core.dataset.providers.s3 import S3Provider
from renku.core.dataset.providers.web import WebProvider
from renku.core.dataset.providers.zenodo import ZenodoProvider
from renku.core.session.docker import DockerSessionProvider
from renku.core.session.renkulab import RenkulabSessionProvider
from renku.core.workflow.converters.cwl import CWLExporter
from renku.core.workflow.converters.renku import RenkuWorkflowFileExporter
from renku.core.workflow.parser.renku import RenkuWorkflowFileParser
from renku.core.workflow.providers.cwltool import CWLToolProvider
from renku.core.workflow.providers.local import LocalWorkflowProvider

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import ProviderApi
    from renku.core.interface.workflow_file_parser import IWorkflowFileParser
    from renku.domain_model.session import ISessionProvider
    from renku.domain_model.workflow.converters import IWorkflowConverter
    from renku.domain_model.workflow.provider import IWorkflowProvider

__all__: List[str] = []

session_providers: "List[Type[ISessionProvider]]" = [DockerSessionProvider, RenkulabSessionProvider]
workflow_exporters: "List[Type[IWorkflowConverter]]" = [CWLExporter, RenkuWorkflowFileExporter]
workflow_providers: "List[Type[IWorkflowProvider]]" = [CWLToolProvider, LocalWorkflowProvider]
dataset_providers: "List[Type[ProviderApi]]" = [
    AzureProvider,
    DataverseProvider,
    ExternalProvider,
    LocalProvider,
    GitProvider,
    OLOSProvider,
    RenkuProvider,
    S3Provider,
    WebProvider,
    ZenodoProvider,
]
workflow_file_parsers: "List[Type[IWorkflowFileParser]]" = [RenkuWorkflowFileParser]

try:
    from renku.core.workflow.providers.toil import ToilProvider

    workflow_providers.append(ToilProvider)
except ImportError:
    pass



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku dataset gateway interface."""

from abc import ABC
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from renku.domain_model.dataset import Dataset, DatasetTag


class IDatasetGateway(ABC):
    """Interface for the DatasetGateway."""

    def get_by_id(self, id: str) -> "Dataset":
        """Get a dataset by id."""
        raise NotImplementedError

    def get_by_slug(self, slug: str) -> Optional["Dataset"]:
        """Get a dataset by slug."""
        raise NotImplementedError

    def get_all_active_datasets(self) -> List["Dataset"]:
        """Get all datasets."""
        raise NotImplementedError

    def get_provenance_tails(self) -> List["Dataset"]:
        """Return the provenance for all datasets."""
        raise NotImplementedError

    def get_all_tags(self, dataset: "Dataset") -> List["DatasetTag"]:
        """Return the list of all tags for a dataset."""
        raise NotImplementedError

    def add_tag(self, dataset: "Dataset", tag: "DatasetTag"):
        """Add a tag from a dataset."""
        raise NotImplementedError

    def remove_tag(self, dataset: "Dataset", tag: "DatasetTag"):
        """Remove a tag from a dataset."""
        raise NotImplementedError

    def add_or_remove(self, dataset: "Dataset") -> None:
        """Add or remove a dataset."""
        raise NotImplementedError



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Datasets JSON-LD schemes."""

from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import DateTimeList, JsonLDSchema, Nested, Uri, fields, oa, prov, renku, schema
from renku.command.schema.entity import CollectionSchema, EntitySchema
from renku.command.schema.image import ImageObjectSchema
from renku.domain_model.dataset import Dataset, DatasetFile, DatasetTag, Language, RemoteEntity, Url


def dump_dataset_as_jsonld(dataset: Dataset) -> dict:
    """Return JSON-LD representation of a dataset.

    Args:
        dataset(Dataset): The dataset to convert.

    Returns:
        dict: JSON-LD data of dataset.

    """
    return DatasetSchema(flattened=True).dump(dataset)


class UrlSchema(JsonLDSchema):
    """Url schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.URL
        model = Url
        unknown = EXCLUDE

    id = fields.Id(load_default=None)
    url = Uri(schema.url, load_default=None)


class DatasetTagSchema(JsonLDSchema):
    """DatasetTag schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.PublicationEvent
        model = DatasetTag
        unknown = EXCLUDE

    dataset_id = Nested(schema.about, UrlSchema, load_default=None)
    date_created = fields.DateTime(schema.startDate, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    description = fields.String(schema.description, load_default=None)
    id = fields.Id()
    name = fields.String(schema.name)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            obj.date_created = obj.date_created.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class LanguageSchema(JsonLDSchema):
    """Language schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Language
        model = Language
        unknown = EXCLUDE

    alternate_name = fields.String(schema.alternateName)
    id = fields.Id(load_default=None)
    name = fields.String(schema.name)


class RemoteEntitySchema(JsonLDSchema):
    """RemoteEntity schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity]
        model = RemoteEntity
        unknown = EXCLUDE

    checksum = fields.String(renku.checksum)
    id = fields.Id()
    path = fields.String(prov.atLocation)
    url = fields.String(schema.url)


class DatasetFileSchema(JsonLDSchema):
    """DatasetFile schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity, schema.DigitalDocument]
        model = DatasetFile
        unknown = EXCLUDE

    based_on = Nested(schema.isBasedOn, RemoteEntitySchema, load_default=None)
    date_added = DateTimeList(schema.dateCreated, format="iso", extra_formats=("%Y-%m-%d",))
    date_removed = fields.DateTime(prov.invalidatedAtTime, load_default=None, format="iso")
    entity = Nested(prov.entity, [EntitySchema, CollectionSchema])
    id = fields.Id()
    is_external = fields.Boolean(renku.external, load_default=False)
    source = fields.String(renku.source, load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.date_added = obj.date_added.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class DatasetSchema(JsonLDSchema):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity, schema.Dataset]
        model = Dataset
        unknown = EXCLUDE

    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    creators = Nested(schema.creator, PersonSchema, many=True)
    date_created = fields.DateTime(schema.dateCreated, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    date_modified = fields.DateTime(schema.dateModified, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    date_removed = fields.DateTime(prov.invalidatedAtTime, load_default=None, format="iso")
    date_published = fields.DateTime(
        schema.datePublished, load_default=None, format="%Y-%m-%d", extra_formats=("iso", "%Y-%m-%dT%H:%M:%S")
    )
    derived_from = Nested(prov.wasDerivedFrom, UrlSchema, load_default=None)
    description = fields.String(schema.description, load_default=None)
    dataset_files = Nested(schema.hasPart, DatasetFileSchema, many=True)
    id = fields.Id(load_default=None)
    identifier = fields.String(schema.identifier)
    images = fields.Nested(schema.image, ImageObjectSchema, load_default=None, many=True)
    in_language = Nested(schema.inLanguage, LanguageSchema, load_default=None)
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)
    license = Uri(schema.license, load_default=None)
    slug = fields.String(renku.slug)
    initial_identifier = fields.String(renku.originalIdentifier)
    project_id = fields.IRI(renku.hasDataset, reverse=True)
    same_as = Nested(schema.sameAs, UrlSchema, load_default=None)
    name = fields.String(schema.name)
    version = fields.String(schema.version, load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            if obj.date_created:
                obj.date_created = obj.date_created.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)
            if obj.date_published:
                obj.date_published = obj.date_published.replace(microsecond=0)
            obj.date_modified = obj.date_modified.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Activity.

``Activity`` represents executed workflows in a Renku project. You can get a
list of all activities in a project by calling its ``list`` method:

.. code-block:: python

    from renku.api import Activity

    activities = Activity.list()

The ``Activity`` class provides a static ``filter`` method that returns a
subset of activities. It can filter activities based on their input, outputs,
parameter names, and parameter values. You can pass a literal value, a list of
values, or a function predicate for each of these fields to filter activities:

.. code-block:: python

    from numbers import Number
    from renku.api import Activity

    # Return activities that use ``path/to/an/input``
    Activity.filter(inputs="path/to/an/input")

    # Return activities that use ``input-1`` or ``input-2`` AND generate
    # output files that their name starts with ``data-``
    Activity.filter(inputs=["input-1", "input-2"], outputs=lambda path: path.startswith("data-"))

    # Return activities that use values between ``0.5`` and ``1.5`` for the
    # parameter ``lr``
    Activity.filter(parameters="lr", values=lambda value: 0.5 <= value <= 1.5 if isinstance(value, Number) else False)

"""

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, List, NamedTuple, Optional, Set, Union

from renku.core import errors
from renku.core.util.os import matches
from renku.domain_model.provenance import activity as core_activity
from renku.domain_model.provenance import parameter as core_parameter
from renku.domain_model.provenance.agent import Person
from renku.domain_model.workflow import plan as core_plan
from renku.ui.api.models.parameter import Input, Output, Parameter, convert_parameter
from renku.ui.api.models.plan import Plan
from renku.ui.api.util import get_activity_gateway, get_plan_gateway

if TYPE_CHECKING:
    from renku.infrastructure.gateway.activity_gateway import ActivityGateway


class Activity:
    """API Activity."""

    def __init__(
        self,
        activity: core_activity.Activity,
        annotations: List,
        generated_outputs: List["GeneratedOutput"],
        ended_at: datetime,
        executed_command: str,
        id: str,
        values: List["FieldValue"],
        plan: Optional[Plan],
        started_at: datetime,
        used_inputs: List["UsedInput"],
        user: str,
    ):
        self._activity: core_activity.Activity = activity
        self._values: Optional[List["FieldValue"]] = values or None
        self._plan: Optional[Plan] = plan
        self.annotations: List = annotations or []
        self.generated_outputs: List["GeneratedOutput"] = generated_outputs or []
        self.ended_at: datetime = ended_at
        self.executed_command: str = executed_command
        self.id: str = id
        self.started_at: datetime = started_at
        self.used_inputs: List["UsedInput"] = used_inputs or []
        self.user: str = user

    @classmethod
    def from_activity(cls, activity: core_activity.Activity) -> "Activity":
        """Create an instance from a core Activity model.

        Args:
            activity(core_activity.Activity): The core activity.

        Returns:
            Activity: An API Activity model.
        """

        def get_user() -> str:
            user = next((a for a in activity.agents if isinstance(a, Person)), None)
            return user.full_identity if user else ""

        return cls(
            activity=activity,
            annotations=activity.annotations,
            generated_outputs=[GeneratedOutput.from_generation(g) for g in activity.generations],
            ended_at=activity.ended_at_time,
            executed_command=" ".join(activity.association.plan.to_argv(with_streams=True)),
            id=activity.id,
            values=[],
            plan=None,
            started_at=activity.started_at_time,
            used_inputs=[UsedInput.from_usage(u) for u in activity.usages],
            user=get_user(),
        )

    @staticmethod
    def list() -> List["Activity"]:
        """List all activities in a project.

        Returns:
            A list of all activities in the supplied project.
        """
        return get_activities()

    @staticmethod
    def filter(
        *,
        inputs: Optional[Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]] = None,
        outputs: Optional[Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]] = None,
        parameters: Optional[Union[str, Iterable[str], Callable[[str], bool]]] = None,
        values: Union[Any, Iterable[Any], Callable[[Any], bool]] = None,
    ) -> List["Activity"]:
        """Return a filtered list of activities in a project.

        Args:
            inputs(Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]): A literal value, list of
                values, or a predicate to filter activities that use an input.
            outputs(Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]): A literal value, list of
                values, or a predicate to filter activities that generate an output.
            parameters(Union[str, Iterable[str], Callable[[str], bool]]): A string, list of strings, or a predicate to
                filter activities based on parameter names that they use.
            values(Union[Any, Iterable[Any], Callable[[Any], bool]]): A literal value, list of values, or a predicate
                to filter activities based on their executed parameter values.

        Returns:
            A list of all activities that match the criteria.
        """
        activity_gateway: Optional["ActivityGateway"] = get_activity_gateway()
        if activity_gateway is None:
            return []

        activities = set(activity_gateway.get_all_activities())

        if inputs is not None:
            filtered = Activity._filter_by_path(
                "get_all_usage_paths", "get_activities_by_usage", path=inputs, activity_gateway=activity_gateway
            )
            activities &= filtered

        if outputs is not None:
            filtered = Activity._filter_by_path(
                "get_all_generation_paths",
                "get_activities_by_generation",
                path=outputs,
                activity_gateway=activity_gateway,
            )
            activities &= filtered

        if parameters is not None or values is not None:
            filtered = Activity._filter_by_parameter(parameters, values, activity_gateway=activity_gateway)
            activities &= filtered

        return [Activity.from_activity(a) for a in activities]

    @staticmethod
    def filter_by_input(path: Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]) -> List["Activity"]:
        """Return a filtered list of activities based on their inputs.

        Args:
            path(Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]): A literal value, list of
                values, or a predicate to filter activities that use an input.

        Returns:
            A list of all activities that match the criteria.
        """
        activities = Activity._filter_by_path("get_all_usage_paths", "get_activities_by_usage", path=path)

        return [Activity.from_activity(a) for a in activities]

    @staticmethod
    def filter_by_output(path: Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]) -> List["Activity"]:
        """Return a filtered list of activities based on their outputs.

        Args:
            path(Union[Path, str, Iterable[Union[Path, str]], Callable[[str], bool]]): A literal value, list of
                values, or a predicate to filter activities that use an input.

        Returns:
            A list of all activities that match the criteria.
        """
        activities = Activity._filter_by_path("get_all_generation_paths", "get_activities_by_generation", path=path)

        return [Activity.from_activity(a) for a in activities]

    @staticmethod
    def _filter_by_path(
        get_all_method: str, get_by_method: str, path, activity_gateway: Optional["ActivityGateway"] = None
    ) -> Set[core_activity.Activity]:
        if activity_gateway is None:
            activity_gateway = get_activity_gateway()
            if activity_gateway is None:
                return set()

        activity_paths = getattr(activity_gateway, get_all_method)()

        if isinstance(path, (Path, str)):
            filtered = [a for a in activity_paths if matches(path=a, pattern=str(path))]
        elif isinstance(path, Iterable):
            filtered = [a for a in activity_paths if any(p for p in path if matches(path=a, pattern=str(p)))]
        elif isinstance(path, Callable):  # type: ignore
            filtered = [a for a in activity_paths if path(a)]
        else:
            raise errors.ParameterError(f"Invalid input/output type: {type(path)}")

        get_activities_by_method = getattr(activity_gateway, get_by_method)
        activities_list = [get_activities_by_method(u) for u in filtered]
        return {a for activities in activities_list for a in activities}

    @staticmethod
    def filter_by_parameter(
        name: Optional[Union[str, Iterable[str], Callable[[str], bool]]] = None,
        value: Optional[Union[Any, Iterable[Any], Callable[[Any], bool]]] = None,
    ) -> List["Activity"]:
        """Return a filtered list of activities based on parameters and their values.

        Args:
            name(Union[str, Iterable[str], Callable[[str], bool]]): A string, list of strings, or a predicate to
                filter activities based on parameter names that they use.
            value(Union[Any, Iterable[Any], Callable[[Any], bool]]): A literal value, list of values, or a predicate
                to filter activities based on their executed parameter values.

        Returns:
            A list of all activities that match the criteria.
        """
        filtered_activities = Activity._filter_by_parameter(name=name, value=value)

        return [Activity.from_activity(a) for a in filtered_activities]

    @staticmethod
    def _filter_by_parameter(
        name: Optional[Union[str, Iterable[str], Callable[[str], bool]]] = None,
        value: Optional[Union[Any, Iterable[Any], Callable[[Any], bool]]] = None,
        activity_gateway: Optional["ActivityGateway"] = None,
    ) -> Set[core_activity.Activity]:
        if activity_gateway is None:
            activity_gateway = get_activity_gateway()
            if activity_gateway is None:
                return set()

        def name_matches(parameter, plan) -> bool:
            if name is None:
                return True

            parameter_name = plan.get_field_by_id(parameter.parameter_id).name

            if isinstance(name, str):
                return name == parameter_name
            elif isinstance(name, Iterable):
                return parameter_name in name
            elif isinstance(name, Callable):  # type: ignore
                return name(parameter_name)

            raise errors.ParameterError(f"Invalid name type: {type(name)}")

        def value_matches(parameter) -> bool:
            if value is None:
                return True
            elif isinstance(value, Iterable):
                return parameter.value in value
            elif isinstance(value, Callable):  # type: ignore
                return value(parameter.value)
            else:
                return value == parameter.value

        activities = activity_gateway.get_all_activities()
        filtered_activities = set()

        for activity in activities:
            for parameter in activity.parameters:
                if name_matches(parameter, plan=activity.association.plan) and value_matches(parameter):
                    filtered_activities.add(activity)
                    break

        return filtered_activities

    def __repr__(self):
        return f"<Activity '{self._activity.association.plan.name}' @ {self.ended_at.isoformat()[:-6]}>"

    @property
    def preceding_activities(self) -> List["Activity"]:
        """Return a list of upstream activities."""
        activity_gateway: Optional["ActivityGateway"] = get_activity_gateway()
        if activity_gateway is None:
            return []

        return [Activity.from_activity(a) for a in activity_gateway.get_upstream_activities(self._activity)]

    @property
    def following_activities(self) -> List["Activity"]:
        """Return a list of downstream activities."""
        activity_gateway: Optional["ActivityGateway"] = get_activity_gateway()
        if activity_gateway is None:
            return []

        return [Activity.from_activity(a) for a in activity_gateway.get_downstream_activities(self._activity)]

    @property
    def values(self) -> List["FieldValue"]:
        """Return list of values for inputs/outputs/parameters that were used in this execution."""
        if self._values is None:
            self._values = [
                FieldValue.from_parameter_value(p, self._activity.association.plan) for p in self._activity.parameters
            ]

        return self._values

    @property
    def plan(self) -> Plan:
        """Activity's base plan."""
        if self._plan is None:
            self._plan = Plan.from_plan(self._activity.association.plan)

        return self._plan


class UsedInput(NamedTuple):
    """Represent an input used in an execution."""

    path: str
    checksum: str

    @classmethod
    def from_usage(cls, usage: core_activity.Usage):
        """Create an instance from core Usage."""
        return cls(path=usage.entity.path, checksum=usage.entity.checksum)

    def __repr__(self):
        return f"<Input '{self.path}'>"


class GeneratedOutput(NamedTuple):
    """Represent an output generated in an execution."""

    path: str
    checksum: str

    @classmethod
    def from_generation(cls, generation: core_activity.Generation):
        """Create an instance from core Generation."""
        return cls(path=generation.entity.path, checksum=generation.entity.checksum)

    def __repr__(self):
        return f"<Output '{self.path}'>"


class FieldValue(NamedTuple):
    """Represent a parameter used in an execution."""

    field: Union[Input, Output, Parameter]
    value: Any

    @classmethod
    def from_parameter_value(cls, parameter: core_parameter.ParameterValue, plan: core_plan.Plan):
        """Create an instance from core ParameterValue."""
        return cls(field=convert_parameter(plan.get_field_by_id(parameter.parameter_id)), value=parameter.value)

    def __repr__(self):
        if isinstance(self.field, Input):
            return f"<Input '{self.field.name}'={self.value}>"
        elif isinstance(self.field, Output):
            return f"<Output '{self.field.name}'={self.value}>"

        return f"<Parameter '{self.field.name}'={self.value}>"


def get_activities(plan_id: Optional[str] = None) -> List[Activity]:
    """Return list of activities that use a plan or one of its predecessor plans."""
    activity_gateway = get_activity_gateway()
    plan_gateway = get_plan_gateway()
    if not activity_gateway or not plan_gateway:
        return []

    activities = activity_gateway.get_all_activities()

    if plan_id:
        ids = {plan_id}
        plan = plan_gateway.get_by_id(plan_id)

        while plan is not None and plan.derived_from is not None:
            ids.add(plan.derived_from)
            plan = plan_gateway.get_by_id(plan.derived_from)

        activities = [a for a in activities if a.association.plan.id in ids]

    return [Activity.from_activity(a) for a in activities]



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version controller."""
import json

from renku.ui.service import config
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.serializers.versions_list import VersionsListResponseRPC
from renku.ui.service.views import result_response


class VersionsListCtrl(ServiceCtrl):
    """Versions list controller."""

    RESPONSE_SERIALIZER = VersionsListResponseRPC()

    def to_response(self):
        """Serialize to service version response."""

        with open(config.METADATA_VERSIONS_LIST, "r") as f:
            return result_response(self.RESPONSE_SERIALIZER, json.load(f))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku graph serializers."""
from marshmallow import Schema, fields, validate

from renku.ui.service.serializers.common import AsyncSchema, GitCommitSHA, MigrateSchema, RemoteRepositorySchema
from renku.ui.service.serializers.rpc import JsonRPCResponse


class GraphExportRequest(AsyncSchema, RemoteRepositorySchema, MigrateSchema, GitCommitSHA):
    """Request schema for dataset list view."""

    callback_url = fields.URL()
    revision = fields.String(load_default="HEAD", allow_none=True)
    format = fields.String(
        load_default="json-ld", validate=validate.OneOf(["json-ld", "nt", "rdf", "dot", "dot-landscape"])
    )


class GraphExportResponse(Schema):
    """Response schema for dataset list view."""

    graph = fields.String()


class GraphExportResponseRPC(JsonRPCResponse):
    """RPC response schema for dataset list view."""

    result = fields.Nested(GraphExportResponse)


class GraphExportCallback(Schema):
    """Callback serializer for graph build."""

    project_url = fields.String()
    commit_id = fields.String()
    type = fields.String(dump_default="RENKU_LOG")


class GraphExportCallbackSuccess(GraphExportCallback):
    """Success callback serializer for graph build."""

    payload = fields.String(dump_default=None, load_default=None)


class GraphExportCallbackError(GraphExportCallback):
    """Error callback serializer for graph build."""

    failure = fields.Dict()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent provenance entities."""

import os.path
from pathlib import Path
from typing import List, Optional, Sequence, Union
from urllib.parse import quote

from renku.infrastructure.immutable import Immutable


class Entity(Immutable):
    """Represent a file."""

    __slots__ = ("checksum", "path")

    path: str
    checksum: str

    def __init__(self, *, checksum: str, id: Optional[str] = None, path: Union[Path, str], **kwargs):
        path = str(path)

        assert id is None or isinstance(id, str)
        assert not os.path.isabs(path), f"Entity is being created with absolute path: '{path}'"

        id = id or Entity.generate_id(checksum, path)
        super().__init__(checksum=checksum, id=id, path=path, **kwargs)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Entity):
            return False
        return self.checksum == other.checksum and self.path == other.path

    def __hash__(self):
        return hash((self.checksum, self.path))

    def __repr__(self) -> str:
        return f"{self.path}#{self.checksum}"

    @staticmethod
    def generate_id(checksum: str, path: Union[Path, str]) -> str:
        """Generate an Entity identifier."""
        quoted_path = quote(str(path).strip("/"))

        return f"/entities/{checksum}/{quoted_path}"


class Collection(Entity):
    """Represent a directory with files."""

    __slots__ = ("members",)

    members: Sequence[Entity]

    def __init__(
        self, *, checksum: str, id: Optional[str] = None, path: Union[Path, str], members: Optional[List[Entity]] = None
    ):
        members_tuple = tuple(members) if members else ()
        super().__init__(checksum=checksum, id=id, path=path, members=members_tuple)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Slots and Immutable classes."""

import threading
import weakref
from typing import Tuple, cast


class Slots:
    """An immutable class.

    Subclasses are supposed to use __slots__ to define their members.
    """

    __slots__ = ("__weakref__",)
    __all_slots__: Tuple[str, ...] = tuple()

    def __new__(cls, *args, **kwargs):
        """Create and return an empty instance of the class."""
        if not cls.__all_slots__:
            cls.__all_slots__ = cast(Tuple[str, ...], cls._get_all_slots())

        return object.__new__(cls)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

    @classmethod
    def make_instance(cls, **kwargs):
        """Instantiate from the given parameters."""
        return cls(**kwargs)

    def __getstate__(self):
        return {name: getattr(self, name, None) for name in self.__class__.__all_slots__ if name != "__weakref__"}

    def __setstate__(self, state):
        new_attributes = set(self.__all_slots__) - set(state)
        for name in new_attributes:
            if name != "__weakref__":
                object.__setattr__(self, name, None)

        for name, value in state.items():
            object.__setattr__(self, name, value)

    @classmethod
    def _get_all_slots(cls):
        all_slots = set()
        for klass in cls.mro():
            if not hasattr(klass, "__slots__"):
                continue
            slots = {klass.__slots__} if isinstance(klass.__slots__, str) else set(klass.__slots__)
            all_slots.update(slots)
        return tuple(all_slots)


class Immutable(Slots):
    """An immutable class that its instances can be cached and reused.

    Immutable subclasses should only contain immutable members. They must call super().__init__(...) to initialize their
    instances. They should not redefine `id` attribute.

    NOTE: Immutable objects must not be modified during the whole provenance and not just only during the object's
    lifetime. These is because we cache these objects and return a single instance if an object with the same `id`
    exists. For example, a DatasetFile object is not immutable across the whole provenance because once it gets removed
    its `date_removed` attribute is set which make the object different from a previous version. As a rule of thumb, an
    object can be immutable if all of its attributes values appear in its id.
    """

    __slots__ = ("id",)

    _local = threading.local()

    @classmethod
    def make_instance(cls, instance):
        """Return a cached instance if available otherwise create an instance from the given parameters."""
        if getattr(cls._local, "cache", None) is None:
            cls._local.cache = weakref.WeakValueDictionary()

        id = getattr(instance, "id", None)
        existing_instance = cls._local.cache.get(id)
        if existing_instance:
            return existing_instance

        if id is not None:
            cls._local.cache[id] = instance

        return instance

    def __setattr__(self, name, value):
        if name != "__weakref__":
            raise TypeError(f"Cannot modify an immutable class {self} {self.__class__}")

        object.__setattr__(self, name, value)

    @property
    def __name__(self):
        return self.__class__.__name__


class DynamicProxy:
    """A proxy class to allow adding dynamic fields to slots/immutable classes."""

    def __init__(self, subject, update=True):
        self._subject = subject
        self._update = update

    def __getattr__(self, name):
        return getattr(self._subject, name)

    def __getattribute__(self, name: str):
        if name == "__class__":
            # NOTE: Makes isinstance() checks work with proxies
            return object.__getattribute__(self._subject, "__class__")

        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name == "_subject" or not hasattr(self._subject, name):
            super().__setattr__(name, value)
        else:
            if not self._update:
                raise ValueError(f"Cannot set attribute '{name}' on {self._subject}.")
            setattr(self._subject, name, value)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tabular format helper functions."""

from collections import OrderedDict
from operator import attrgetter

from renku.core import errors


def tabulate(collection, columns, columns_mapping, columns_alignments=None, sort=True, reverse=False) -> str:
    """Format collection with a tabular output.

    Args:
        collection: Collection to format.
        columns: Columns to show.
        columns_mapping: Mapping of collection fields to columns.
        columns_alignments: Column alignment (Default value = None).
        sort: Whether to sort by first column or not (Default value = True).
        reverse: Whether to sort in reverse (Default value = False).
    """
    from renku.core.util.tabulate import tabulate as to_table

    if not columns:
        raise errors.ParameterError("Columns cannot be empty.")

    columns = [c.lower().strip() for c in columns.split(",") if c]

    headers, alignments = _make_headers(columns, columns_mapping, columns_alignments)

    # Sort based on the first requested field
    if sort:
        try:
            attr = list(headers.keys())[0]
            getter = attrgetter(attr)
            collection = sorted(collection, key=getter, reverse=reverse)
        except TypeError:
            pass

    alignments = alignments if collection else None  # To avoid a tabulate bug

    return to_table(collection, headers=headers, colalign=alignments, disable_numparse=True)


def _make_headers(columns, columns_mapping, columns_alignments):
    columns_alignments = columns_alignments or {}
    headers = OrderedDict()
    alignments = []
    for column in columns:
        if column not in columns_mapping:
            raise errors.ParameterError(
                'Invalid column name: "{}".\nPossible values: {}'.format(column, ", ".join(columns_mapping))
            )
        name, display_name = columns_mapping.get(column)
        headers[name] = display_name

        alignment = columns_alignments.get(column, "left")
        alignments.append(alignment)

    return headers, alignments



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder for repository."""

from pathlib import Path
from typing import List, Optional, Union

from renku.command.command_builder.command import Command, CommandResult, check_finalized
from renku.core import errors
from renku.core.git import ensure_clean
from renku.domain_model.project_context import project_context


class Commit(Command):
    """Builder for commands that create a commit."""

    HOOK_ORDER = 4

    def __init__(
        self,
        builder: Command,
        message: Optional[str] = None,
        commit_if_empty: Optional[bool] = False,
        raise_if_empty: Optional[bool] = False,
        commit_only: Optional[Union[str, List[Union[str, Path]]]] = None,
        skip_staging: bool = False,
        skip_dirty_checks: bool = False,
    ) -> None:
        """__init__ of Commit.

        Args:
            message (str): The commit message. Auto-generated if left empty (Default value = None).
            commit_if_empty (bool): Whether to commit if there are no modified files (Default value = None).
            raise_if_empty (bool): Whether to raise an exception if there are no modified files (Default value = None).
            commit_only (bool): Only commit the supplied paths (Default value = None).
            skip_staging(bool): Don't commit staged files.
            skip_dirty_checks(bool): Don't check if paths are dirty or staged.
        """
        self._builder = builder
        self._message = message
        self._commit_if_empty = commit_if_empty
        self._raise_if_empty = raise_if_empty
        self._commit_filter_paths = commit_only
        self._skip_staging: bool = skip_staging
        self._skip_dirty_checks: bool = skip_dirty_checks

    def _pre_hook(self, builder: Command, context: dict, *args, **kwargs) -> None:
        """Hook to create a commit transaction.

        Args:
            builder(Command): The current ``CommandBuilder``.
            context(dict): The current context object.
        """
        from renku.core.util.git import prepare_commit

        if "stack" not in context:
            raise ValueError("Commit builder needs a stack to be set.")

        self.diff_before = prepare_commit(
            repository=project_context.repository,
            commit_only=self._commit_filter_paths,
            skip_staging=self._skip_staging,
            skip_dirty_checks=self._skip_dirty_checks,
        )

    def _post_hook(self, builder: Command, context: dict, result: CommandResult, *args, **kwargs):
        """Hook that commits changes.

        Args:
            builder(Command):The current ``CommandBuilder``.
            context(dict): The current context object.
            result(CommandResult): The result of the command execution.
        """
        from renku.core.util.git import finalize_commit

        if result.error:
            # TODO: Cleanup repo
            return

        try:
            finalize_commit(
                diff_before=self.diff_before,
                repository=project_context.repository,
                transaction_id=project_context.transaction_id,
                commit_only=self._commit_filter_paths,
                commit_empty=self._commit_if_empty,
                raise_if_empty=self._raise_if_empty,
                commit_message=self._message,
                skip_staging=self._skip_staging,
            )
        except errors.RenkuException as e:
            result.error = e

    @check_finalized
    def build(self) -> Command:
        """Build the command.

        Returns:
            Command: Finalized version of this command.
        """
        self._builder.add_pre_hook(self.HOOK_ORDER, self._pre_hook)
        self._builder.add_post_hook(self.HOOK_ORDER, self._post_hook)

        return self._builder.build()

    @check_finalized
    def with_commit_message(self, message: str) -> Command:
        """Set a new commit message.

        Args:
            message(str): Commit message to set.

        Returns:
            Command: This command with commit message applied.
        """
        self._message = message

        return self


class RequireClean(Command):
    """Builder to check if repo is clean."""

    HOOK_ORDER = 4

    def __init__(self, builder: Command) -> None:
        """__init__ of RequireClean."""
        self._builder = builder

    def _pre_hook(self, builder: Command, context: dict, *args, **kwargs) -> None:
        """Check if repo is clean.

        Args:
            builder(Command): Current ``CommandBuilder``.
            context(dict): Current context.
        """
        if not project_context.has_context():
            raise ValueError("Commit builder needs a ProjectContext to be set.")

        ensure_clean(ignore_std_streams=not builder._track_std_streams)

    @check_finalized
    def build(self) -> Command:
        """Build the command.

        Returns:
            Command: Finalized version of this command.
        """
        self._builder.add_pre_hook(self.HOOK_ORDER, self._pre_hook)

        return self._builder.build()


class Isolation(Command):
    """Builder to run a command in git isolation."""

    HOOK_ORDER = 3

    def __init__(
        self,
        builder: Command,
    ) -> None:
        """__init__ of Commit."""
        self._builder = builder

    def _injection_pre_hook(self, builder: Command, context: dict, *args, **kwargs) -> None:
        """Hook to setup dependency injection for commit transaction.

        Args:
            builder(Command): Current ``CommandBuilder``.
            context(dict): Current context.
        """
        from renku.core.git import prepare_worktree

        if not project_context.has_context():
            raise ValueError("Commit builder needs a ProjectContext to be set.")

        _, self.isolation, self.path, self.branch_name = prepare_worktree(path=None, branch_name=None, commit=None)

    def _post_hook(self, builder: Command, context: dict, result: CommandResult, *args, **kwargs):
        """Hook that commits changes.

        Args:
            builder(Command): Current ``CommandBuilder``.
            context(dict): Current context.
        """
        from renku.core.git import finalize_worktree

        try:
            finalize_worktree(
                isolation=self.isolation,
                path=self.path,
                branch_name=self.branch_name,
                delete=True,
                new_branch=True,
                exception=result.error,
            )
        except errors.RenkuException as e:
            if not result.error:
                result.error = e

    @check_finalized
    def build(self) -> Command:
        """Build the command.

        Returns:
            Command: Finalized version of this command.
        """
        self._builder.add_injection_pre_hook(self.HOOK_ORDER, self._injection_pre_hook)
        self._builder.add_post_hook(self.HOOK_ORDER, self._post_hook)

        return self._builder.build()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service views."""
from flask import jsonify
from marshmallow import Schema

from renku.ui.service.config import SVC_ERROR_GENERIC
from renku.ui.service.serializers.rpc import JsonRPCResponse


def result_response(serializer: Schema, data):
    """Construct flask response."""
    return jsonify(serializer.dump({"result": data}))


def error_response(serviceError):
    """Construct error response."""
    error = {}
    error["code"] = getattr(serviceError, "code", SVC_ERROR_GENERIC)
    error["userMessage"] = getattr(serviceError, "userMessage", "Unexpected exception")
    error["devMessage"] = getattr(serviceError, "devMessage", error["userMessage"])
    if hasattr(serviceError, "userReference"):
        error["userReference"] = serviceError.userReference
    if hasattr(serviceError, "devReference"):
        error["devReference"] = serviceError.devReference
    if hasattr(serviceError, "sentry"):
        error["sentry"] = serviceError.sentry

    return jsonify(JsonRPCResponse().dump({"error": error}))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for renku run customization."""
import pluggy

hookspec = pluggy.HookspecMarker("renku")


@hookspec
def plan_annotations(plan):
    """Plugin Hook to add ``Annotation`` entry list to a ``Plan``.

    Run when a Plan is created by ``renku run``.

    Args:
        plan: A ``Plan`` object to get annotations for.

    Returns:
        A list of ``renku.domain_model.provenance.annotation.Annotation``
        objects.

    """
    pass


@hookspec
def activity_annotations(activity):
    """Plugin Hook to add ``Annotation`` entry list to a ``Activity``.

    Run when creating an activity from a ``renku run``.

    Args:
        activity: An ``Activity`` object to get annotations for.

    Returns:
        A list of ``renku.domain_model.provenance.annotation.Annotation``
        objects.

    """
    pass


@hookspec
def pre_run(tool):
    """Plugin Hook that gets called at the start of a ``renku run`` call.

    Can be used to setup plugins that get executed during the run.

    Args:
        tool (``PlanFactory``): The plan factory that captured the run.
    """
    pass



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Merge strategies."""

import os
import shutil
import traceback
from json import JSONDecodeError
from pathlib import Path
from tempfile import mkdtemp
from typing import List, NamedTuple, Optional, Union, cast

from BTrees.OOBTree import BTree, Bucket, TreeSet
from deepdiff import DeepDiff
from persistent import Persistent
from persistent.list import PersistentList
from zc.relation.catalog import Catalog

from renku.core import errors
from renku.core.constant import DATABASE_PATH, RENKU_HOME
from renku.core.util import communication
from renku.domain_model.dataset import Dataset, Url
from renku.domain_model.project import Project
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.plan import AbstractPlan
from renku.infrastructure.database import Database, Index
from renku.infrastructure.repository import Repository
from renku.version import __version__


class RemoteEntry(NamedTuple):
    """Reference to an entry in a database on a separate branch."""

    reference: str
    database: Database
    path: Path
    repository: Repository


class GitMerger:
    """Git metadata merger."""

    def merge(self, local: Path, remote: Path, base: Path) -> None:
        """Merge two renku metadata entries together."""
        repository = project_context.repository
        self.remote_entries: List[RemoteEntry] = []

        self._setup_worktrees(repository)

        merged = False
        self.local_database = project_context.database

        try:
            local_object = self.local_database.get_from_path(str(project_context.path / local))
            try:
                base_object: Optional[Persistent] = self.local_database.get_from_path(str(project_context.path / base))
            except (errors.ObjectNotFoundError, JSONDecodeError):
                base_object = None

            for entry in self.remote_entries:
                # NOTE: Loop through all remote merge branches (Octo merge) and try to merge them
                try:
                    self.remote_database = entry.database
                    remote_object = self.remote_database.get_from_path(str(project_context.path / remote))

                    # NOTE: treat merge result as new local for subsequent merges
                    local_object = self.merge_objects(local_object, remote_object, base_object)
                    merged = True
                except errors.ObjectNotFoundError:
                    continue
        finally:
            # NOTE: cleanup worktrees
            for entry in self.remote_entries:
                repository.remove_worktree(entry.path)
                shutil.rmtree(entry.path, ignore_errors=True)

        if not merged:
            raise errors.MetadataMergeError("Couldn't merge metadata: remote object not found in merge branches.")

        self.local_database.persist_to_path(local_object, local)

    def _setup_worktrees(self, repository):
        """Setup git worktrees for the remote branches."""

        # NOTE: Get remote branches
        remote_branches = [os.environ[k] for k in os.environ.keys() if k.startswith("GITHEAD")]

        database_path = Path(RENKU_HOME) / DATABASE_PATH

        for remote_branch in remote_branches:
            # NOTE: Create a new shallow worktree for each remote branch, could be several in case of an octo merge
            worktree_path = Path(mkdtemp())
            repository.create_worktree(worktree_path, reference=remote_branch, checkout=False)
            try:
                remote_repository = Repository(worktree_path)
                remote_repository.checkout(sparse=[database_path])

                self.remote_entries.append(
                    RemoteEntry(
                        remote_branch,
                        Database.from_path(worktree_path / database_path),
                        worktree_path,
                        remote_repository,
                    )
                )
            except Exception:
                exc_str = traceback.format_exc(limit=None, chain=True)
                if "No such file or directory" in exc_str and "Unable to create" in exc_str:
                    communication.error(
                        "Error when trying to sparse checkout worktree. This is likely due to using an old version of "
                        "git. Please try with a newer version."
                    )
                # NOTE: cleanup worktree
                try:
                    repository.remove_worktree(worktree_path)
                except Exception:  # nosec
                    pass
                raise

    def merge_objects(self, local: Persistent, remote: Persistent, base: Optional[Persistent]) -> Persistent:
        """Merge two database objects."""
        if not isinstance(local, type(remote)):
            raise errors.MetadataMergeError(f"Cannot merge {local} and {remote}: disparate types.")
        if isinstance(local, (BTree, Index, Bucket)):
            return self.merge_btrees(local, remote)
        elif isinstance(local, TreeSet):  # type: ignore[unreachable]
            return self.merge_treesets(local, remote)
        elif isinstance(local, Catalog):
            return self.merge_catalogs(local, remote)
        elif isinstance(local, Project):
            return self.merge_projects(local, remote, cast(Optional[Project], base))
        else:
            raise errors.MetadataMergeError(
                f"Cannot merge {local} and {remote}: type not supported for automated merge."
            )

    def merge_btrees(
        self, local: Union[BTree, Index, Bucket], remote: Union[BTree, Index, Bucket]
    ) -> Union[BTree, Index, Bucket]:
        """Merge two BTrees."""
        local_key_ids = {k: getattr(v, "_p_oid", None) for k, v in local.items()}
        remote_key_ids = {k: getattr(v, "_p_oid", None) for k, v in remote.items()}

        common_modified_keys = [k for k, i in local_key_ids.items() if remote_key_ids.get(k, i) != i]
        new_remote_keys = [k for k in remote_key_ids.keys() if k not in local_key_ids]

        for new_remote in new_remote_keys:
            # NOTE: New entries in remote, we can just copy them over
            local[new_remote] = remote[new_remote]

        for common_key in common_modified_keys:
            # NOTE: Merge conflicts!
            local_object = local[common_key]
            remote_object = remote[common_key]

            comparison = self._compare_objects(local_object, remote_object)

            if comparison is not None:
                local[common_key] = comparison
                continue

            local_object._p_activate()
            remote_object._p_activate()
            diff = DeepDiff(local_object, remote_object, exclude_types=[Database])
            pretty_diff = diff.pretty().replace("Value of root.", "local.")
            pretty_diff = "\n".join(f"\t{line}" for line in pretty_diff.splitlines())
            entry_type = str(type(local_object)).split(".")[-1][:-2]

            action = communication.prompt(
                "Merge conflict detected:\n"
                f"{common_key} ({entry_type}) modified in local and remote branch.\n"
                f"Changes between local and remote:\n{pretty_diff}\n"
                "Which do you want to keep?\n[l]ocal, [r]emote, [a]bort:",
                default="a",
            )

            if action == "r":
                local[common_key] = remote_object
            elif action == "a":
                raise errors.MetadataMergeError("Merge aborted")
            elif action != "l":
                raise errors.MetadataMergeError(f"Invalid merge option selected: {action}")

        return local

    def merge_treesets(self, local: TreeSet, remote: TreeSet) -> TreeSet:
        """Merge two TreeSets."""
        local.update([e for e in remote if e not in local])
        return local

    def merge_indices(self, local: Index, remote: Index) -> Index:
        """Merge two Indices."""
        local_key_ids = {k: getattr(v, "_p_oid", None) for k, v in local.items()}
        remote_key_ids = {k: getattr(v, "_p_oid", None) for k, v in remote.items()}

        common_modified_keys = [k for k, i in local_key_ids.items() if remote_key_ids.get(k, i) != i]
        new_remote_keys = [k for k in remote_key_ids.keys() if k not in local_key_ids]

        for new_remote in new_remote_keys:
            # NOTE: New entries in remote, we can just copy them over
            local.add(remote.get(new_remote))

        for common_key in common_modified_keys:
            # NOTE: Merge conflicts!
            local_object = local.get(common_key)
            remote_object = remote.get(common_key)

            comparison = self._compare_objects(local_object, remote_object)

            if comparison is not None:
                local[common_key] = comparison
                continue

            local_object._p_activate()
            remote_object._p_activate()
            diff = DeepDiff(local_object, remote_object, exclude_types=[Database])
            pretty_diff = diff.pretty().replace("Value of root.", "local.")
            pretty_diff = "\n".join(f"\t{line}" for line in pretty_diff.splitlines())
            entry_type = str(type(local.get(common_key))).split(".")[-1][:-2]

            action = communication.prompt(
                "Merge conflict detected:\n"
                f"{common_key} ({entry_type}) modified in local and remote branch.\n"
                f"Changes between local and remote:\n{pretty_diff}\n"
                "Which do you want to keep?\n[l]ocal, [r]emote, [a]bort:",
                default="a",
            )

            if action == "r":
                local.pop(common_key)
                local.add(remote_object)
            elif action == "a":
                raise errors.MetadataMergeError("Merge aborted")
            elif action != "l":
                raise errors.MetadataMergeError(f"Invalid merge option selected: {action}")

        return local

    def merge_catalogs(self, local: Catalog, remote: Catalog) -> Catalog:
        """Merge two Catalogs."""
        for key, value in remote._EMPTY_name_TO_relcount_relset.items():
            if key not in local._EMPTY_name_TO_relcount_relset:
                local._EMPTY_name_TO_relcount_relset[key] = value

        for key, value in remote._name_TO_mapping.items():
            if key not in local._name_TO_mapping:
                local._name_TO_mapping[key] = value
                continue
            for subkey, subvalue in value.items():
                if subkey not in local._name_TO_mapping[key]:
                    local._name_TO_mapping[key][subkey] = subvalue

        for key, value in remote._reltoken_name_TO_objtokenset.items():
            if key not in local._reltoken_name_TO_objtokenset:
                local._reltoken_name_TO_objtokenset[key] = value

        return local

    def merge_projects(self, local: Project, remote: Project, base: Optional[Project]) -> Project:
        """Merge two Project entries."""

        local_changed = (
            base is None
            or local.keywords != base.keywords
            or local.description != base.description
            or local.annotations != base.annotations
        )
        local_template_changed = base is None or local.template_metadata != base.template_metadata
        remote_changed = (
            base is None
            or remote.keywords != base.keywords
            or remote.description != base.description
            or remote.annotations != base.annotations
        )
        remote_template_changed = base is None or remote.template_metadata != base.template_metadata

        if (local_changed or local_template_changed) and not remote_changed and not remote_template_changed:
            return local
        elif not local_changed and not local_template_changed and (remote_changed or remote_template_changed):
            return remote

        if local_changed or remote_changed:
            # NOTE: Merge keywords
            if local.keywords != remote.keywords:
                if base is None:
                    local.keywords = list(set(local.keywords) | set(remote.keywords))
                elif local.keywords != base.keywords and remote.keywords != base.keywords:
                    removed = (set(base.keywords) - set(local.keywords)) | (set(base.keywords) - set(remote.keywords))
                    added = (set(local.keywords) - set(base.keywords)) | (set(remote.keywords) - set(base.keywords))
                    existing = set(base.keywords) - removed
                    local.keywords = list(added | existing)
                elif remote.keywords != base.keywords:
                    local.keywords = remote.keywords

            # NOTE: Merge description
            if local.description != remote.description:
                if base is None or (local.description != base.description and remote.description != base.description):
                    local.description = communication.prompt(
                        f"Project description was modified in local and remote branch.\n"
                        f"local: {local.description}\nremote: {remote.description}\nEnter merged description: ",
                        default=local.description,
                    )
                elif remote.description != base.description:
                    local.description = remote.description

            # NOTE: Merge annotations
            local.annotations = list(set(local.annotations) | set(remote.annotations))

        # NOTE: Merge versions
        if int(remote.version) > int(local.version):
            local.version = remote.version

        local.agent_version = __version__

        # NOTE: Merge template data
        if local_template_changed and remote_template_changed:
            # NOTE: Merge conflicts!
            action = communication.prompt(
                "Merge conflict detected:\n Project template modified/update in both remote and local branch.\n"
                f"local: {local.template_metadata.template_source}@{local.template_metadata.template_ref}:"
                f"{local.template_metadata.template_id}, "
                "version {local.template_metadata.template_version}\n"
                f"remote: {remote.template_metadata.template_source}@{remote.template_metadata.template_ref}:"
                f"{remote.template_metadata.template_id}, "
                "version {remote.template_metadata.template_version}\n"
                "Which do you want to keep?\n[l]ocal, [r]emote, [a]bort:",
                default="a",
            )

            if action == "r":
                local.template_metadata = remote.template_metadata
            elif action == "a":
                raise errors.MetadataMergeError("Merge aborted")
            elif action != "l":
                raise errors.MetadataMergeError(f"Invalid merge option selected: {action}")
        elif remote_template_changed:
            local.template_metadata = remote.template_metadata

        return local

    def _compare_objects(self, local_object: Persistent, remote_object: Persistent) -> Optional[Persistent]:
        """Compare two objects and return the newer/relevant one, if possible."""
        if local_object == remote_object or getattr(local_object, "id", object()) == getattr(
            remote_object, "id", object()
        ):
            # NOTE: Objects are the same, nothing to do
            return local_object

        if isinstance(local_object, Dataset) and local_object.is_derivation():
            if self._is_dataset_derived_from(local_object, cast(Dataset, remote_object), self.local_database):
                return local_object
        if isinstance(remote_object, Dataset) and remote_object.is_derivation():
            if self._is_dataset_derived_from(remote_object, cast(Dataset, local_object), self.remote_database):
                return remote_object
        if isinstance(local_object, AbstractPlan) and local_object.is_derivation():
            if self._is_plan_derived_from(local_object, cast(AbstractPlan, remote_object), self.local_database):
                return local_object
        if isinstance(remote_object, AbstractPlan) and remote_object.is_derivation():
            if self._is_plan_derived_from(remote_object, cast(AbstractPlan, local_object), self.remote_database):
                return remote_object
        if isinstance(local_object, list) and isinstance(remote_object, list):
            local_object.extend(r for r in remote_object if r not in local_object)
            return local_object
        if isinstance(local_object, PersistentList) and isinstance(remote_object, PersistentList):
            local_object.extend(r for r in remote_object if r not in local_object)
            return local_object

        return None

    def _is_dataset_derived_from(self, potential_child: Dataset, potential_parent: Dataset, database: Database) -> bool:
        """Check if a dataset is a derivation of another dataset."""
        parent = potential_child
        while parent.is_derivation():
            parent = database.get_by_id(cast(Url, parent.derived_from).value)
            if parent.id == potential_parent.id:
                return True

        return False

    def _is_plan_derived_from(
        self, potential_child: AbstractPlan, potential_parent: AbstractPlan, database: Database
    ) -> bool:
        """Check if a dataset is a derivation of another dataset."""
        parent = potential_child
        while parent.is_derivation():
            parent = database.get_by_id(cast(str, parent.derived_from))
            if parent.id == potential_parent.id:
                return True

        return False
